import type_enforced
import sys

import pydoc


@type_enforced.Enforcer
class Foo:
    @classmethod
    def add(self, a: int, b: int) -> int:
        """
        Add Docs Here
        """
        return a + b

    @staticmethod
    def subtract(a: int, b: int) -> int:
        """
        Subtract Docs Here
        """
        return a - b

    def multiply(self, a: int, b: int) -> int:
        """
        Multiply Docs Here
        """
        return a * b


docstring = pydoc.render_doc(Foo)

docstring_checks = ["Multiply Docs Here", "Subtract Docs Here", "Add Docs Here"]

if any([i not in docstring for i in docstring_checks]):
    print("test_class_07.py failed")
else:
    print("test_class_07.py passed")



import type_enforced
from typing import Literal


@type_enforced.Enforcer
def my_fn(a: Literal["bar"] | int) -> None:
    pass


success = True
# TODO: In the next major version, add support to enforce combinations of literals as types
#       and not constraints.
# try:
#     my_fn("bar")  # Passes
#     my_fn(1)  # Passes
# except:
#     success = False


if success:
    print(f"test_fn_18.py passed")
else:
    print(f"test_fn_18.py failed")



import type_enforced


@type_enforced.Enforcer
class my_class:
    def my_fn(self, a: int, b: [int, str] = 2, c: int = 3) -> None:
        return None

    def my_fn_args(self, a: int, *args, b: [int, str] = 2, c: int = 3) -> None:
        return None

    def my_fn_kwargs(
        sellf, a: int, b: [int, str] = 2, c: int = 3, **kwargs
    ) -> None:
        return None

    def my_fn_args_kwargs(
        self, a: int, *args, b: [int, str] = 2, c: int = 3, **kwargs
    ) -> None:
        return None


obj = my_class()
success = True
success_type_error = True
try:
    obj.my_fn(a=1, b=2, c=3)  # No Error
    obj.my_fn_args(a=1, b=2, c=3)  # No Error
    obj.my_fn_kwargs(a=1, b=2, c=3)  # No Error
    obj.my_fn_args_kwargs(a=1, b=2, c=3)  # No Error
except:
    success = False

try:
    obj.my_fn(a="a", b=2, c=3)  # Error (a can only accept int)
    success = False
except Exception as e:
    if type(e) != TypeError:
        success_type_error = False

try:
    obj.my_fn_args(a="a", b=2, c=3)  # Error (a can only accept int)
    success = False
except Exception as e:
    if type(e) != TypeError:
        success_type_error = False

try:
    obj.my_fn_kwargs(a="a", b=2, c=3)  # Error (a can only accept int)
    success = False
except Exception as e:
    if type(e) != TypeError:
        success_type_error = False

try:
    obj.my_fn_args_kwargs(a="a", b=2, c=3)  # Error (a can only accept int)
    success = False
except Exception as e:
    if type(e) != TypeError:
        success_type_error = False


if success and success_type_error:
    print("test_class_08.py passed")
else:
    print("test_class_08.py failed")



import type_enforced
from typing import List, Set, Dict, Tuple


@type_enforced.Enforcer
def my_fn(
    a: List[int], b: Set[str], c: Dict[str, int], d: Tuple[str, int]
) -> None:
    return None


success_1 = True
try:
    my_fn(
        a=[1, 2, 3], b={"a", "b", "c"}, c={"a": 1, "b": 2, "c": 3}, d=("a", 1)
    )
except:
    success_1 = False

success_2 = False
try:
    my_fn(
        a=[1, 2, 3],
        b={"a", "b", "c"},
        c={"a": 1, "b": 2, "c": 3},
        d=("a", 1.5, 2),
    )  # Error (d can only accept Tuple[str, int])
except Exception as e:
    if "Type mismatch" in str(e):
        success_2 = True

if success_1 and success_2:
    print("test_fn_09.py passed")
else:
    print("test_fn_09.py failed")



import type_enforced
from typing import Callable


# Create a generator to return items in a list
def get_sum_till_now(items):
    total = 0
    for item in items:
        total += item
        yield total


def my_function(x):
    return x


class MyClass:
    def __init__(self):
        pass

    def my_method(self, x):
        return x

    @staticmethod
    def my_static_method(x):
        return x

    @classmethod
    def my_class_method(self, x):
        return x


@type_enforced.Enforcer
def foo(out: str) -> Callable:
    if out == "sum":
        return sum
    elif out == "generator":
        return get_sum_till_now(range(10))
    elif out == "lambda":
        return lambda x: x
    elif out == "function":
        return my_function
    elif out == "method":
        x = MyClass()
        return x.my_method
    elif out == "staticmethod":
        return MyClass.my_static_method
    elif out == "classmethod":
        x = MyClass()
        return x.my_class_method
    else:
        return None


passed = True
try:
    foo("sum")
    foo("generator")
    foo("lambda")
    foo("function")
    foo("method")
    foo("staticmethod")
    foo("classmethod")
except:
    passed = False

try:
    foo("other")
    passed = False
except:
    pass

if passed:
    print("test_fn_14.py passed")
else:
    print("test_fn_14.py failed")



import type_enforced


@type_enforced.Enforcer
def my_fn_enforced(a: int):
    pass


@type_enforced.Enforcer
def my_fn(a):
    pass


# Ensure that functions without annotations are not type enforced to prevent unnecessary overhead
success = isinstance(
    my_fn_enforced, type_enforced.FunctionMethodEnforcer
) and not isinstance(my_fn, type_enforced.FunctionMethodEnforcer)

if success:
    print("test_fn_03.py passed")
else:
    print("test_fn_03.py failed")



import type_enforced


@type_enforced.Enforcer
def my_fn(x: [list[str] | list[int] | list[list[int]]]) -> None:
    pass


def inv_my_fn(x: [list[int] | list[str] | list[list[int]]]) -> None:
    pass


success = True

try:
    my_fn([1, 2, 3])  # Passes
    my_fn(["a", "b", "c"])  # Passes
    my_fn([[1, 2], [3, 4]])  # Passes

    inv_my_fn([1, 2, 3])  # Passes
    inv_my_fn(["a", "b", "c"])  # Passes
    inv_my_fn([[1, 2], [3, 4]])  # Passes
except:
    success = False

# TODO: In the next major version, add support to enforce
#       multiple different types of nested lists but not mixed types
# try:
#     my_fn([1, "a", 3])
#     success = False
# except:
#     pass


if success:
    print(f"test_fn_17.py passed")
else:
    print(f"test_fn_17.py failed")



import type_enforced
from type_enforced.utils import Constraint, GenericConstraint

CustomConstraint = GenericConstraint(
    {
        "in_rgb": lambda x: x in ["red", "green", "blue"],
    }
)


@type_enforced.Enforcer()
def positive_int_test(value: [int, Constraint(ge=0)]) -> bool:
    return True


@type_enforced.Enforcer()
def positive_float_test(value: [int, float, Constraint(ge=0)]) -> bool:
    return True


@type_enforced.Enforcer()
def running_str_test(value: [str, Constraint(pattern=r".*running.*")]) -> bool:
    return True


@type_enforced.Enforcer()
def custom_constraint_test(value: [str, CustomConstraint]) -> bool:
    return True


success = True
try:
    positive_int_test(0)
except TypeError as err:
    success = False

try:
    positive_int_test(-1)
    success = False
except TypeError:
    pass

try:
    positive_int_test("Hello There")
    success = False
except TypeError:
    pass

try:
    positive_float_test(0.1)
except TypeError:
    success = False

try:
    positive_float_test(-0.99)
    success = False
except TypeError:
    pass

try:
    positive_float_test("Hello There")
    success = False
except TypeError:
    pass

try:
    running_str_test(0)
    success = False
except TypeError:
    pass

try:
    running_str_test("this is a stopped status")
    success = False
except TypeError:
    pass

try:
    running_str_test("this is running status")
except TypeError:
    success = False

try:
    custom_constraint_test("red")
except TypeError:
    success = False

try:
    custom_constraint_test("yellow")
    success = False
except TypeError:
    pass


if success:
    print(f"test_fn_16.py passed")
else:
    print(f"test_fn_16.py failed")



import type_enforced
from typing import Sized


@type_enforced.Enforcer
def my_fn(a: Sized) -> int:
    return a.__len__()


success_1 = True
try:
    my_fn(a=["a"])  # No Error
    my_fn(a={"a": 1})  # No Error
    my_fn(a=(1,))  # No Error
    my_fn(a={1})  # No Error
    my_fn(a="a")  # No Error
    my_fn(a=memoryview(b"abc"))  # No Error
    my_fn(a=b"abc")  # No Error
    my_fn(a=bytearray(b"abc"))  # No Error
    my_fn(a=range(1))  # No Error
except:
    success_1 = False

success_2 = False
try:
    my_fn(a=1)
except Exception as e:
    if "Type mismatch" in str(e):
        success_2 = True

if success_1 and success_2:
    print("test_fn_12.py passed")
else:
    print("test_fn_12.py failed")
